/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.channel;

import java.io.IOException;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketException;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.Push;
import org.objectweb.dream.message.codec.MessageCodec;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Abstract implementation of a ChannelOut component.
 */
public abstract class AbstractTCPChannelOutImpl extends AbstractComponent
    implements
      Push,
      ChannelOutAttributeController
{

  protected int            cnxRetry = 5;

  protected MessageManager messageManagerItf;
  protected MessageCodec   messageCodecItf;

  // ---------------------------------------------------------------------------
  // Implementation of ChannelOutAttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see ChannelOutAttributeController#getConnectionRetry()
   */
  public int getConnectionRetry()
  {
    return cnxRetry;
  }

  /**
   * @see ChannelOutAttributeController#setConnectionRetry(int)
   */
  public void setConnectionRetry(int retry)
  {
    cnxRetry = retry;
  }

  // ---------------------------------------------------------------------------
  // Implementation of BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{MessageManager.ITF_NAME, MessageCodec.ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
    else if (clientItfName.equals(MessageCodec.ITF_NAME))
    {
      messageCodecItf = (MessageCodec) serverItf;
    }
  }

  // ---------------------------------------------------------------------------
  // Utility methods
  // ---------------------------------------------------------------------------

  /**
   * This method connects and returns a socket.
   * 
   * @param addr the IP to connect to
   * @param port the port number
   * @return a socket
   * @throws IOException if after many attempt, the connection failed, an
   *           IOException is thrown
   */
  protected Socket connectSocket(InetAddress addr, int port) throws IOException
  {
    for (int i = 0;; i++)
    {
      try
      {
        logger.log(BasicLevel.DEBUG, "try to connect");
        Socket sock = new Socket(addr, port);
        setSocketOption(sock);
        logger.log(BasicLevel.DEBUG, "connected");
        return sock;
      }
      catch (IOException exc)
      {
        logger.log(BasicLevel.WARN, "connection failed on address " + addr);
        if (i > cnxRetry)
        {
          throw exc;
        }
        try
        {
          Thread.sleep(i * 250);
        }
        catch (InterruptedException ignored)
        {
          // ignored
        }
      }
    }
  }

  /**
   * This method close a socket.
   * 
   * @param out a socket
   * @throws IOException if after many attempt, the connection failed, an
   *           IOException is thrown
   */
  protected void closeSocket(Socket out) throws IOException
  {
    out.close();
  }

  /**
   * Configures this socket using the socket options established for this
   * factory. It may be overloaded in subclass, in order to handle particular
   * subclasses of sockets
   * 
   * @param sock the socket.
   * @exception SocketException for networking errors
   */
  protected void setSocketOption(Socket sock) throws SocketException
  {
    // Don't use TCP data coalescing - ie Nagle's algorithm
    sock.setTcpNoDelay(true);
    // Read operation will block indefinitely until requested data arrives
    sock.setSoTimeout(0);
    // Set Linger-on-Close timeout.
    sock.setSoLinger(true, 60);
  }

}