/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.channel;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.net.ConnectException;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.Collections;

import org.objectweb.dream.InitializationException;
import org.objectweb.dream.control.activity.Util;
import org.objectweb.dream.control.activity.task.AbstractTask;
import org.objectweb.dream.control.activity.task.Task;
import org.objectweb.dream.message.Message;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Basic implementation of a TCP ChannelIn Component.
 */
public class BasicTCPChannelInImpl extends AbstractTCPChannelInImpl
{

  /** The {@link ServerSocket}used by this component */
  protected ServerSocket listen;

  protected Task         channelInTask = new ChannelInTask();

  // ---------------------------------------------------------------------------
  // Inner task class implementation
  // ---------------------------------------------------------------------------

  protected class ChannelInTask extends AbstractTask
  {

    /** Constructor */
    public ChannelInTask()
    {
      super("ChannelInTask");
    }

    /**
     * @see Task#execute(Object)
     */
    public Object execute(Object hints) throws InterruptedException
    {
      Socket socket;
      ObjectInputStream ois = null;
      try
      {
        if (listen == null || listen.isClosed())
        {
          logger.log(BasicLevel.DEBUG, "ServerSocket closed, create new one");
          listen = createServerSocket(portNum);
        }
        try
        {
          logger.log(BasicLevel.DEBUG, "waiting connection");
          socket = listen.accept();
        }
        catch (SocketException ex)
        {
          logger.log(BasicLevel.WARN, ex);
          return EXECUTE_AGAIN;
        }
        setSocketOption(socket);
        logger.log(BasicLevel.DEBUG, "connected");
        try
        {

          ois = new ObjectInputStream(socket.getInputStream());

          Message msg = messageCodecItf.decode(ois);

          logger.log(BasicLevel.DEBUG, "msg received");
          outPushItf.push(msg, null);
          // then send the acknowledge.
          logger.log(BasicLevel.DEBUG, "send ack");

          sendAck(socket);

        }
        finally
        {
          try
          {
            ois.close();
          }
          catch (Exception exc)
          {
            logger.log(BasicLevel.WARN, exc);
          }
          ois = null;
          try
          {
            socket.close();
          }
          catch (Exception exc)
          {
            logger.log(BasicLevel.WARN, exc);
          }
          socket = null;
        }

      }
      catch (ConnectException e)
      {
        logger.log(BasicLevel.DEBUG, e);
      }
      catch (IOException exc)
      {
        logger.log(BasicLevel.ERROR, "Output/Input Stream failed", exc);
      }
      catch (Exception exc)
      {
        logger.log(BasicLevel.ERROR, "Exception during ", exc);
      }
      return EXECUTE_AGAIN;
    }
  }

  // ---------------------------------------------------------------------------
  // Overriding of initComponent method
  // ---------------------------------------------------------------------------

  protected void initComponent(Component componentItf)
      throws InitializationException
  {
    super.initComponent(componentItf);
    try
    {
      Util.addTask(componentItf, channelInTask, Collections.EMPTY_MAP);
    }
    catch (Exception e)
    {
      throw new InitializationException(e, componentItf, "Can't add task");
    }
  }

  // ---------------------------------------------------------------------------
  // Overriding of LifeCycleController methods
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.LifeCycleController#stopFc()
   */
  public synchronized void stopFc() throws IllegalLifeCycleException
  {
    try
    {
      listen.close();
    }
    catch (IOException e)
    {
      // ignore
    }
    listen = null;
    super.stopFc();
  }
}