/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.control.lifecycle;

import org.objectweb.dream.control.activity.task.IllegalTaskException;
import org.objectweb.dream.control.activity.task.NoSuchTaskException;
import org.objectweb.dream.control.activity.task.Task;
import org.objectweb.dream.control.activity.task.TaskActivationController;
import org.objectweb.dream.control.activity.task.TaskController;
import org.objectweb.dream.control.activity.task.TaskStoppedListener;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.julia.control.lifecycle.ChainedIllegalLifeCycleException;
import org.objectweb.fractal.julia.control.lifecycle.LifeCycleCoordinator;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * This mixin can be applied to a lifecycle controller in order to start the
 * activities of the component to which the controller belongs.
 * <p>
 * <p>
 * <b>Requirements </b>
 * <ul>
 * <li>This mixin must be used with the
 * {@link org.objectweb.fractal.julia.UseComponentMixin }mixin.</li>
 * <li>This mixin must be used with the
 * {@link org.objectweb.dream.control.activity.task.UseTaskControllerMixin}
 * mixin.</li>
 * <li>This mixin must be used with the
 * {@link org.objectweb.dream.control.activity.task.UseTaskActivationControllerMixin}
 * mixin.</li>
 * <li>This mixin must be used with the {@link LoggableLifeCycleMixin }mixin.
 * </li>
 * <li>This mixin must be used with the
 * {@link org.objectweb.fractal.julia.control.lifecycle.BasicLifeCycleControllerMixin}
 * lifecycle controller implementation.</li>
 * </ul>
 */
public abstract class ActivityLifeCycleMixin implements TaskStoppedListener
{

  protected int                  fcNbStoppingTask;
  protected LifeCycleCoordinator fcActivityCoordinator;

  // ---------------------------------------------------------------------------
  // Private constructor
  // ---------------------------------------------------------------------------

  private ActivityLifeCycleMixin()
  {
  }

  // ---------------------------------------------------------------------------
  // Fields and methods added and overriden by the mixin class
  // ---------------------------------------------------------------------------

  /**
   * Calls super implementation and starts internal activities in the started
   * state.
   * 
   * @see LifeCycleCoordinator#setFcStarted()
   */
  public synchronized boolean setFcStarted() throws IllegalLifeCycleException
  {
    if (!_super_setFcStarted())
    {
      return false;
    }

    // restart tasks
    Task[] tasks = _this_weaveableTC.getTasks();
    for (int i = 0; i < tasks.length; i++)
    {
      if (_this_weaveableLCCLogger.isLoggable(BasicLevel.DEBUG))
      {
        _this_weaveableLCCLogger.log(BasicLevel.DEBUG, "activate task "
            + tasks[i]);
      }
      try
      {
        _this_weaveableTAC.activateTask(tasks[i]);
      }
      catch (NoSuchTaskException e)
      {
        // cannot happend
        Error.bug(_this_weaveableLCCLogger, e);
      }
      catch (IllegalTaskException e)
      {
        throw new ChainedIllegalLifeCycleException(e, _this_weaveableC,
            "Error while registering task");
      }
    }
    return true;
  }

  /**
   * Calls super implementation and looks for this component's coordinator
   * controller.
   * 
   * @see LifeCycleCoordinator#setFcStopping(LifeCycleCoordinator)
   */
  public void setFcStopping(final LifeCycleCoordinator coordinator)
      throws IllegalLifeCycleException
  {
    Task[] tasks = _this_weaveableTC.getTasks();
    synchronized (this)
    {
      fcNbStoppingTask += tasks.length;

      fcActivityCoordinator = coordinator;
      for (int i = 0; i < tasks.length; i++)
      {
        if (_this_weaveableLCCLogger.isLoggable(BasicLevel.DEBUG))
        {
          _this_weaveableLCCLogger.log(BasicLevel.DEBUG, "interrupt task "
              + tasks[i]);
        }
        try
        {
          _this_weaveableTAC.deactivateTask(tasks[i], this);
        }
        catch (NoSuchTaskException e)
        {
          // cannot happend
          Error.bug(_this_weaveableLCCLogger, e);
        }
        catch (IllegalTaskException e)
        {
          throw new ChainedIllegalLifeCycleException(e, _this_weaveableC,
              "Error while unregistering task");
        }
      }
    }
    _super_setFcStopping(coordinator);
  }

  /**
   * Returns <code>true</code> If the component has no executing task and
   * super implementation returns <code>true</code>.
   * 
   * @return <code>true</code> If the component is inactive
   */
  public boolean checkInactivity()
  {
    return fcNbStoppingTask == 0 && _super_checkInactivity();
  }

  //---------------------------------------------------------------------------
  // Implementation of TaskStoppedListener interface
  //---------------------------------------------------------------------------

  /**
   * @see TaskStoppedListener#taskStopped(Task)
   */
  public void taskStopped(Task task)
  {
    synchronized (this)
    {
      if (_this_weaveableLCCLogger.isLoggable(BasicLevel.DEBUG))
      {
        _this_weaveableLCCLogger.log(BasicLevel.DEBUG, "task stopped" + task);
      }
      fcNbStoppingTask--;
      _this_isInactivated(fcActivityCoordinator);
    }
  }

  // ---------------------------------------------------------------------------
  // Fields and methods required by the mixin class in the base class
  // ---------------------------------------------------------------------------

  /**
   * The <tt>weaveableC</tt> field required by this mixin. This field is
   * supposed to reference the {@link Component}interface of the component to
   * which this controller object belongs.
   */
  public Component                _this_weaveableC;

  /**
   * The <tt>weaveableLCCLogger</tt> field required by this mixin. This field
   * is supposed to reference the {@link Logger }of this controller.
   */
  public Logger                   _this_weaveableLCCLogger;

  /**
   * The <tt>weaveableTC</tt> field required by this mixin. This field is
   * supposed to reference the {@link TaskController }interface of the
   * component to which this controller object belongs.
   */
  public TaskController           _this_weaveableTC;

  /**
   * The <code>weaveableTAC</code> field required by this mixin. This field is
   * supposed to reference the {@link TaskActivationController }interface of
   * the component to which this controller object belongs.
   */
  public TaskActivationController _this_weaveableTAC;

  /**
   * The {@link LifeCycleCoordinator#setFcStarted() }method overriden by this
   * mixin.
   * 
   * @see org.objectweb.fractal.julia.control.lifecycle.BasicLifeCycleControllerMixin
   */
  public abstract boolean _super_setFcStarted()
      throws IllegalLifeCycleException;

  /**
   * The {@link LifeCycleCoordinator#setFcStopping(LifeCycleCoordinator) }
   * method overriden by this mixin.
   * 
   * @see org.objectweb.fractal.julia.control.lifecycle.BasicLifeCycleControllerMixin
   */
  public abstract void _super_setFcStopping(
      final LifeCycleCoordinator coordinator) throws IllegalLifeCycleException;

  /**
   * The <code>checkInactivity</code> method overriden by this mixin. This
   * method is supposed to work as this
   * {@link SimpleLifeCycleControllerMixin#checkInactivity() }method.
   * 
   * @see SimpleLifeCycleControllerMixin#checkInactivity()
   */
  public abstract boolean _super_checkInactivity();

  /**
   * The <code>isInactivated</code> method used by this mixin. This method is
   * supposed to work as this
   * {@link SimpleLifeCycleControllerMixin#isInactivated }method.
   * 
   * @see SimpleLifeCycleControllerMixin#isInactivated
   */
  public abstract boolean _this_isInactivated(LifeCycleCoordinator coordinator);
}