/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Vivien Quema
 */

package org.objectweb.dream.message;

import java.util.Iterator;

/**
 * Base interface of every messages manipulated by Dream components. A message
 * is composed of named chunks and a list of sub messages. More precisely, A
 * <code>Message</code> is a <i>naming context </i> for its chunks. This means
 * that a message can not have two chunks with the same name. In addition a
 * message can have two chunks with the same type since they have different
 * names.
 * 
 * @see org.objectweb.dream.message.Chunk
 * @see org.objectweb.dream.message.MessageType
 */
public interface Message
{

  /**
   * An empty <code>Message</code> array constant. This constant should be
   * used by classes requiring an empty <code>Message</code> array.
   */
  Message[] EMPTY_MESSAGE_ARRAY = new Message[0];

  /**
   * Returns the id of the message manager that created this message.
   * 
   * @return the id of the message manager that created this message.
   */
  short getMessageManagerId();

  /**
   * Sets the id of the message manager that created this message.
   * 
   * @param id the id of the message manager that created this message.
   */
  void setMessageManagerId(short id);

  /**
   * Returns a chunk of this message, or <code>null</code> if this message
   * doesn't have a chunk with the specified name.
   * 
   * @param name the name of the chunk, as specified in the message type.
   * @return a chunk of this message.
   */
  Object getChunk(String name);

  /**
   * Returns an <code>Iterator</code> that iterate over the sub messages of
   * this message.
   * 
   * @return an iterator on the sub messages.
   */
  Iterator getSubMessageIterator();

  /**
   * Returns an (eventually empty) array containing the sub messages of this
   * message.
   * 
   * @return an array containing the sub messages of this message.
   */
  Message[] getSubMessages();

  /**
   * Returns the type of this message.
   * 
   * @return the type of this message.
   */
  MessageType getMessageType();

  /**
   * Transferts the state of the message's chunks into the specified new
   * instance.
   * <p>
   * <i>Note: </i> the given new isntance contains all the chunks contained in
   * this message.
   * 
   * @param newInstance the new instance of message.
   */
  void transfertChunkStates(Message newInstance);

  /**
   * Recycles message instance. Called by message manager when the message is
   * deleted.
   */
  void recycle();
}