/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s):
 */

package org.objectweb.dream.protocol.causality;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.ExtensibleMessage;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.chunk.ArrowChunk;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Incomming message sorter component. This component can be seen as a passive
 * Push/Push queue.
 */
public class InMessageSorterImpl extends AbstractComponent
    implements
      Push,
      CausalityTransformerAttributeController
{

  private MatrixClock    matrixClockItf;
  private Push           outPushItf;
  private MessageManager messageManagerItf;

  private List           waitingToDeliver;
  private String         arrowChunkName;
  private String         causalityChunkName;

  /**
   * Default constructor
   */
  public InMessageSorterImpl()
  {
    waitingToDeliver = new LinkedList();
  }

  /**
   * @see Push#push(Message, Map)
   */
  public void push(Message message, Map context) throws PushException
  {
    ArrowChunk arrowChunk = (ArrowChunk) message.getChunk(arrowChunkName);

    if (arrowChunk == null)
    {
      logger.log(BasicLevel.ERROR, "Unable to find arrow chunk named "
          + arrowChunkName + ". The message is dropped");
      return;
    }

    CausalityChunk csltChunk = (CausalityChunk) message
        .getChunk(causalityChunkName);

    if (csltChunk == null)
    {
      logger.log(BasicLevel.ERROR, "Unable to find causality chunk named "
          + causalityChunkName + ". The message is dropped");
    }

    int todo = matrixClockItf.testRecvMatrix(csltChunk.getCausalityStamp(),
        arrowChunk.getArrowFrom(), arrowChunk.getArrowTo());
    if (todo == MatrixClock.DELIVER)
    {
      deliverMessage(message);

      // scan waiting to deliver message if they can be delivred.
      boolean loop = true;
      while (loop)
      {
        loop = false;
        Iterator iter = waitingToDeliver.iterator();
        while (iter.hasNext())
        {
          Message msg = (Message) iter.next();
          ArrowChunk arrowChunk1 = (ArrowChunk) message
              .getChunk(arrowChunkName);
          CausalityChunk csltChunk1 = (CausalityChunk) message
              .getChunk(causalityChunkName);

          if (matrixClockItf.testRecvMatrix(csltChunk1.getCausalityStamp(),
              arrowChunk1.getArrowFrom(), arrowChunk.getArrowTo()) == MatrixClock.DELIVER)
          {

            deliverMessage(msg);
            iter.remove();
            // Matrix clock has changed, we have to rescan the list.
            loop = true;
            break;
          }
        }
      }
    }
    else if (todo == MatrixClock.WAIT_TO_DELIVER)
    {
      // insert in the waitting list
      waitingToDeliver.add(message);
      if (logger.isLoggable(BasicLevel.DEBUG))
      {
        logger.log(BasicLevel.DEBUG, "block message " + message);
      }
    }
    else
    {
      // message already delivred.
      if (logger.isLoggable(BasicLevel.DEBUG))
      {
        logger.log(BasicLevel.DEBUG, "message already delivred " + message);
      }
    }
  }

  private void deliverMessage(Message message) throws PushException
  {
    // try to remove causality chunk
    if (message instanceof ExtensibleMessage)
    {
      Object causalityChunk = ((ExtensibleMessage) message)
          .removeChunk(causalityChunkName);
      messageManagerItf.deleteChunk(causalityChunk);
    }

    outPushItf.push(message, null);
    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "deliver message " + message);
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of CausalityTransformerAttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see CausalityTransformerAttributeController#getArrowChunkName()
   */
  public String getArrowChunkName()
  {
    return arrowChunkName;
  }

  /**
   * @see CausalityTransformerAttributeController#setArrowChunkName(String)
   */
  public void setArrowChunkName(String arrowChunkName)
  {
    this.arrowChunkName = arrowChunkName;
  }

  /**
   * @see CausalityTransformerAttributeController#getCausalityChunkName()
   */
  public String getCausalityChunkName()
  {
    return causalityChunkName;
  }

  /**
   * @see CausalityTransformerAttributeController#setCausalityChunkName(String)
   */
  public void setCausalityChunkName(String causalityChunkName)
  {
    this.causalityChunkName = causalityChunkName;
  }

  // ---------------------------------------------------------------------------
  // Implementation of BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{Push.OUT_PUSH_ITF_NAME,
        MatrixClock.MATRIX_CLOCK_ITF_NAME, MessageManager.ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(MatrixClock.MATRIX_CLOCK_ITF_NAME))
    {
      matrixClockItf = (MatrixClock) serverItf;
    }
    else if (clientItfName.equals(Push.OUT_PUSH_ITF_NAME))
    {
      outPushItf = (Push) serverItf;
    }
    else if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
  }

}