/**
 * DREAM
 * Copyright (C) 2003 2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s):
 */

package org.objectweb.dream.protocol.causality;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

/**
 * Matrix update elements
 */
public class Update implements Externalizable
{

  /** Matrix clock line element. */
  private transient short  l;

  /** Matrix clock column element. */
  private transient short  c;

  /** Matrix clock value. */
  private transient int    stamp;

  /** pointer on next element of update list. */
  private transient Update next;

  /**
   * Default contructor. Defines for serialization.
   */
  public Update()
  {
  }

  /**
   * Creates an element in a new empty list.
   * 
   * @param l Matrix line.
   * @param c Matrix column.
   * @param stamp Element stamp.
   */
  public Update(short l, short c, int stamp)
  {
    super();
    this.l = l;
    this.c = c;
    this.stamp = stamp;
  }

  /**
   * Creates an element and links it after the head.
   * 
   * @param l Matrix line.
   * @param c Matrix column.
   * @param s Element stamp.
   * @param list The head list element.
   */
  public Update(short l, short c, int s, Update list)
  {
    this.l = l;
    this.c = c;
    this.stamp = s;
    if (list != null)
    {
      this.next = list.next;
      list.next = this;
    }
    else
    {
      this.next = null;
    }
  }

  /**
   * @return Matrix clock line element.
   */
  public short getLine()
  {
    return l;
  }

  /**
   * @return Matrix clock column element.
   */
  public short getColumn()
  {
    return c;
  }

  /**
   * @return Matrix clock value.
   */
  public int getStamp()
  {
    return stamp;
  }

  /**
   * @return pointer on next element of update list.
   */
  public Update getNext()
  {
    return next;
  }

  /**
   * @see java.lang.Object#toString()
   */
  public String toString()
  {
    StringBuffer sb = new StringBuffer();
    sb.append('(');
    Update u = this;
    while (u != null)
    {
      u.toStringBuffer(sb);
      u = u.next;
    }
    sb.append(')');
    return sb.toString();
  }

  private StringBuffer toStringBuffer(StringBuffer sb)
  {
    sb.append('(');
    sb.append(l).append(", ");
    sb.append(c).append(", ");
    sb.append(stamp).append(')');
    return sb;
  }

  /**
   * @see java.io.Externalizable#readExternal(java.io.ObjectInput)
   */
  public void readExternal(ObjectInput in) throws IOException,
      ClassNotFoundException
  {
    this.l = in.readShort();
    this.c = in.readShort();
    this.stamp = in.readInt();

    short l;
    while ((l = in.readShort()) != -1)
    {
      alloc(l, in.readShort(), in.readInt(), this);
    }
  }

  /**
   * @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
   */
  public void writeExternal(ObjectOutput out) throws IOException
  {
    Update update = this;
    while (update != null)
    {
      out.writeShort(update.l);
      out.writeShort(update.c);
      out.writeInt(update.stamp);
      Update nextUpdate = update.next;
      unalloc(update);
      update = nextUpdate;
    }
    out.writeShort(-1);
  }

  /**
   * @see java.io.Externalizable#readExternal(ObjectInput)
   */
  public static Update readUpdate(ObjectInput in) throws IOException,
      ClassNotFoundException
  {
    Update u = alloc((short) 0, (short) 0, 0, null);
    u.readExternal(in);
    return u;
  }

  private static Update[] pool = new Update[30];

  private static int      elementCount;

  static synchronized Update alloc(short l, short c, int s, Update list)
  {
    if (elementCount == 0)
    {
      return new Update(l, c, s, list);
    }
    elementCount -= 1;
    Update update = pool[elementCount];

    update.l = l;
    update.c = c;
    update.stamp = s;
    if (list != null)
    {
      update.next = list.next;
      list.next = update;
    }
    else
    {
      update.next = null;
    }

    pool[elementCount] = null;
    return update;

  }

  static synchronized void unalloc(Update update)
  {
    if (elementCount == pool.length)
    {
      return;
    }
    pool[elementCount] = update;
    update.next = null;
    elementCount += 1;
  }
}